#!/usr/bin/env python

import sys
import logging
import math
import numpy as np
from numpy import linalg as LA

from CodeEntropy.poseidon.extractData.generalFunctions import com, MOI, vector, distance
from collections import defaultdict
nested_dict = lambda: defaultdict(nested_dict) 
        #create nested dict in one go





def calculateFTMatrix(all_data, dimensions):
    '''
    Consider force/torque calculations as discussed in Ali 2019 paper.
    UA level and molecule level.
    '''

    resid_list = []
    for x in range(0, len(all_data)):
        atom = all_data[x]
        if atom.mass > 1.1:
            if atom.resid not in resid_list: #iterate though resids once
                resid_list.append(atom.resid)
                all_molecule_atoms = [] #list of heavy atoms and Hs
                UAs_list = [] #list of lists [HA + Hs]

                for num in atom.molecule_atomNums: #atom nums in molecule
                    UA = all_data[num]

                    all_molecule_atoms.append(UA)
                    heavy_bonded = []
                    H_bonded = []

                    for bonded_atom in UA.bonded_to_atom_num:
                        bonded = all_data[bonded_atom]
                        if bonded.mass > 1.1 and bonded.atom_num != \
                                UA.atom_num:
                            heavy_bonded.append(bonded)

                        if bonded.mass < 1.1 and bonded.atom_num != \
                                UA.atom_num:
                            H_bonded.append(bonded)
                            all_molecule_atoms.append(bonded)
                            
                    UA_atoms_list = [UA] + H_bonded
                    UAs_list.append(UA_atoms_list)
                    bonded_atoms_list = [UA] + heavy_bonded + H_bonded
                    UA.bondedUA_H = [len(heavy_bonded), len(H_bonded)]


                molecule_coords = []
                molecule_masses = []
                for a in all_molecule_atoms:
                    molecule_coords.append(a.coords)
                    molecule_masses.append(a.mass)
                molecule_COM = com(molecule_coords, molecule_masses)


                #####****** MOLECULE LEVEL ******
                #if molecule is only one UA with bonded Hs, 
                #this works for one UA molecules
                if len(atom.molecule_atomNums) == 1 \
                        and len(atom.bonded_to_atom_num) > 0:
                    WM_principal_axes, WM_MI_axis = principalAxesMOI(
                            all_data, all_molecule_atoms, 
                            molecule_COM, Hs=True)
                            #use atom list containing Hs for P axes inc. Hs
                    WM_force, WM_torque = rotateFT(all_data, 
                            all_molecule_atoms, 
                            WM_principal_axes, WM_principal_axes, WM_MI_axis, 
                            molecule_COM, molecule_COM, Hs=True)

                    #'''
                    WM_force = np.outer(WM_force, WM_force)
                    WM_torque = np.outer(WM_torque, WM_torque)
                    atom.MweightedForces = np.round(np.divide(WM_force, 4), 3)
                    atom.MweightedTorques = np.round(np.divide(WM_torque, 4), 3)
                    atom.WMprincipalAxis = WM_principal_axes, \
                            WM_MI_axis, molecule_COM

                    ###check this with Richard, do you not halve here? 
                    ###For Jon's code, you do halve here
                    ###even though molecule == UA level
                    atom.UAweightedForces = np.round(np.divide(WM_force, 4), 3)
                    atom.UAweightedTorques = np.round(np.divide(WM_torque, 4), 3)
                    atom.molecule_UA_Fs = np.round(np.divide(WM_force, 4), 3)
                    atom.molecule_UA_Ts = np.round(np.divide(WM_torque, 4), 3)
                    #'''

                    '''
                    atom.MweightedForces = WM_force
                    atom.MweightedTorques = WM_torque
                    atom.UAweightedForces = WM_force
                    atom.UAweightedTorques = WM_torque
                    '''



                #if molecule contains more than one UA
                #forces work, torques work if Hs = True
                if len(atom.molecule_atomNums) > 1:

                    WM_principal_axes, WM_MI_axis = principalAxesMOI(
                            all_data, all_molecule_atoms, 
                            molecule_COM, Hs=False)
                            #only use heavy atom list for P axes calc
                            #we don't consider Hs as they result 
                            #in larger torques as Hs are further away

                    WM_force, WM_torque = rotateFT(all_data, 
                            all_molecule_atoms, 
                            WM_principal_axes, WM_principal_axes, WM_MI_axis, 
                            molecule_COM, molecule_COM, Hs=True)

                    #'''
                    #WM_FT = np.concatenate((WM_force, WM_torque), axis=None)
                    #WM_FT = np.outer(WM_FT, WM_FT)
                    WM_force = np.outer(WM_force, WM_force)
                    WM_torque = np.outer(WM_torque, WM_torque)
                    atom.MweightedForces = np.round(np.divide(WM_force, 4), 3)
                    atom.MweightedTorques = np.round(np.divide(WM_torque, 4), 3)
                    atom.WMprincipalAxis = WM_principal_axes, \
                            WM_MI_axis, molecule_COM

                    atom.molecule_UA_Fs = np.round(np.divide(WM_force, 4), 3)
                    atom.molecule_UA_Ts = np.round(np.divide(WM_torque, 4), 3)
                    #'''

                    #atom.MweightedForces = WM_force
                    #atom.MweightedTorques = WM_torque


                    ##### ****** UNITED-ATOM LEVEL ******
                    ## works!
                    UA_F_list = []
                    UA_T_list = []
                    for UA in UAs_list:
                        XX_principal_axes, UA_MI_axis \
                                = None, None

                        if UA[0].mass > 1.1:
                            bonded_HAs = [UA[0]] #inc itself
                            bonded_Hs = []
                            for b in UA[0].bonded_to_atom_num:
                                bonded = all_data[b]
                                if bonded.mass < 1.1:
                                    bonded_Hs.append(bonded)
                                elif bonded.mass > 1.1:
                                    bonded_HAs.append(bonded)
                                else:
                                    continue


                            if len(bonded_HAs) == 2 and len(bonded_Hs) > 0:
                                #num HAs including itself
                                UA_COM = bonded_HAs[0].coords

                                R = bonded_HAs[0]
                                X1 = bonded_HAs[1]
                                H = bonded_Hs[0]

                                RX1_vector = vector(R.coords, 
                                        X1.coords, dimensions)
                                RX1_dist = distance(R.coords, 
                                        X1.coords, dimensions)
                                Paxis1 = np.divide(RX1_vector, RX1_dist)

                                RH_vector = vector(R.coords, 
                                        H.coords, dimensions)
                                RH_dist = distance(R.coords, 
                                        H.coords, dimensions)
                                RH_vector_norm = np.divide(RH_vector, RH_dist)

                                Paxis2 = np.cross(Paxis1, RH_vector_norm)
                                Paxis3 = np.cross(Paxis1, Paxis2)


                                XX_principal_axes = [Paxis1, Paxis2, Paxis3]

                                UA_MI_axis, XX_principal_axes = \
                                        UA_MOI(all_data, 
                                        UA, UA_COM, XX_principal_axes, 
                                        dimensions, Hs=True) 

                                if len(bonded_Hs) == 1:
                                    MIx, MIy, MIz = UA_MI_axis[0], \
                                            UA_MI_axis[1], UA_MI_axis[2]
                                    if MIx < MIy and MIx < MIz:
                                        MIx = 0
                                    if MIy < MIx and MIy < MIz:
                                        MIy = 0
                                    if MIz < MIx and MIz < MIy:
                                        MIz = 0

                                    UA_MI_axis = [MIx, MIy, MIz]


                            elif len(bonded_HAs) > 2:
                                #num HAs including itself
                                UA_COM = bonded_HAs[0].coords

                                R = bonded_HAs[0]
                                X1 = bonded_HAs[1]
                                X2 = bonded_HAs[2]

                                Paxis1 = np.zeros(3) 
                                    #average of all HA covalent bond vectors
                                for HA in bonded_HAs[1:]:
                                    Paxis1 += vector(R.coords, 
                                            HA.coords, dimensions)
                                    
                                X1X2_vector = vector(X1.coords, 
                                        X2.coords, dimensions)
                                X1X2_dist = distance(X1.coords, 
                                        X2.coords, dimensions)
                                X1X2_norm = np.divide(X1X2_vector, X1X2_dist)
                                Paxis2 = np.cross(X1X2_norm, Paxis1)

                                Paxis3 = np.cross(Paxis1, Paxis2)


                                XX_principal_axes = [Paxis1, Paxis2, Paxis3]

                                UA_MI_axis, XX_principal_axes = \
                                        UA_MOI(all_data, 
                                        UA, UA_COM, XX_principal_axes, 
                                        dimensions, Hs=True)

                            elif len(bonded_HAs) == 2 and len(bonded_Hs) == 0:
                                #num use arbitrary coords for 2nd vector
                                UA_COM = bonded_HAs[0].coords

                                R = bonded_HAs[0]
                                X1 = bonded_HAs[1]

                                RX1_vector = vector(R.coords, 
                                        X1.coords, dimensions)
                                RX1_dist = distance(R.coords, 
                                        X1.coords, dimensions)
                                Paxis1 = np.divide(RX1_vector, RX1_dist)

                                RZ_vector = vector(R.coords, 
                                        [0, 0, 0], dimensions)
                                RZ_dist = distance(R.coords, 
                                        [0, 0, 0], dimensions)
                                RZ_vector_norm = np.divide(RZ_vector, RZ_dist)

                                Paxis2 = np.cross(Paxis1, RZ_vector_norm)
                                Paxis3 = np.cross(Paxis1, Paxis2)


                                XX_principal_axes = [Paxis1, Paxis2, Paxis3]

                                UA_MI_axis, XX_principal_axes = \
                                        UA_MOI(all_data, 
                                        UA, UA_COM, XX_principal_axes, 
                                        dimensions, Hs=True) 

                                if len(bonded_Hs) == 0:
                                    MIx, MIy, MIz = UA_MI_axis[0], \
                                            UA_MI_axis[1], UA_MI_axis[2]
                                    if MIx < MIy and MIx < MIz:
                                        MIx = 0
                                    if MIy < MIx and MIy < MIz:
                                        MIy = 0
                                    if MIz < MIx and MIz < MIy:
                                        MIz = 0

                                    UA_MI_axis = [MIx, MIy, MIz]

                            else:
                                continue



                                
                        if UA[0].mass < 1.1:
                            print ('Error: no HA in UA')

                        UA_force, UA_torque = rotateFT(all_data, UA, 
                                WM_principal_axes, 
                                XX_principal_axes, UA_MI_axis, 
                                molecule_COM, UA_COM, Hs=True)
                                #torques use HA com, 
                                #forces use molecule COM

                        UA_F_list += [UA_force]
                        UA_T_list += [UA_torque]


                        #'''
                        UA_force = np.outer(UA_force, UA_force)
                        UA_torque = np.outer(UA_torque, UA_torque)
                        UA[0].UAweightedForces = np.round(
                                np.divide(UA_force, 4), 3)
                        ##UA[0].UAweightedForces = UA_force
                                #2019 paper, dont halve forces
                        UA[0].UAweightedTorques = np.round(
                                np.divide(UA_torque, 4), 3)
                        #'''

                        #UA[0].UAweightedForces = UA_force
                        #UA[0].UAweightedTorques = UA_torque


                    #dont halve forces in molecule UAs (2019 paper),
                    #divide by 4 to compare with jons output
                    molecule_UA_Fs = np.concatenate(UA_F_list, axis=None)
                    molecule_UA_Fs = np.outer(molecule_UA_Fs, molecule_UA_Fs)
                    #atom.molecule_UA_Fs = np.divide(molecule_UA_Fs, 4)
                    atom.molecule_UA_Fs = np.round(molecule_UA_Fs)


                    #halve torques in molecule UAs
                    molecule_UA_Ts = np.concatenate(UA_T_list, axis=None)
                    molecule_UA_Ts = np.outer(molecule_UA_Ts, molecule_UA_Ts)
                    atom.molecule_UA_Ts = np.round(np.divide(molecule_UA_Ts, 4), 3)






                #if molecule is monoatomic (one UA and no bonded Hs)
                if len(atom.molecule_atomNums) == 1 \
                        and len(atom.bonded_to_atom_num) == 0:
                    mass_sqrt = float(atom.mass ** 0.5)
                    forces_sqrt = [float(atom.forces[0]) / mass_sqrt,
                            float(atom.forces[1]) / mass_sqrt,
                            float(atom.forces[2]) / mass_sqrt]
                    forces_sqrt = np.sort(forces_sqrt)

                    #'''
                    UA_force = np.outer(forces_sqrt, forces_sqrt)
                    UA_torque = np.outer([0, 0, 0], [0, 0, 0])
                    atom.UAweightedForces = np.round(np.divide(UA_force, 4), 3)
                    atom.UAweightedTorques = np.round(UA_torque, 3) #zero
                    atom.MweightedForces = np.round(np.divide(UA_force, 4), 3)
                    atom.MweightedTorques = np.round(UA_torque, 3) #zero
                    atom.WMprincipalAxis = [[0, 0, 0], [0, 0, 0], [0, 0, 0]], \
                            [0, 0, 0], [0, 0, 0]

                    atom.molecule_UA_Fs = np.round(np.divide(UA_force, 4), 3)
                    atom.molecule_UA_Ts = np.round(np.divide(UA_torque, 4), 3)
                    #'''

                    '''
                    atom.UAweightedForces = forces_sqrt
                    atom.UAweightedTorques = [0, 0, 0] #zero
                    atom.MweightedForces = forces_sqrt
                    atom.MweightedTorques = [0, 0, 0] #zero
                    '''





def UA_MOI(all_data, bonded_atoms_list, COM, PIaxes, dimensions, Hs):
    '''
    Get MOI axis for UA level, where eigenvalues and vectors are
    not used
    '''


    coord_list = []
    mass_list = []

    for atom in bonded_atoms_list:
        if Hs == True:
            #print (atom.atom_name)
            coord_list.append(atom.coords)
            mass_list.append(atom.mass)
        elif Hs == False:
            if atom.mass > 1.1:
                UA_mass = 0 #add mass of HA and bonded Hs
                #print (atom.atom_name)
                UA_mass += atom.mass
                coord_list.append(atom.coords) #only include coords of UA
                for h in atom.bonded_to_atom_num:
                    H = all_data[h]
                    if H.mass < 1.1:
                        UA_mass += H.mass
                    else:
                        continue
                mass_list.append(UA_mass) 
                        #include mass of Hs on heavy atom mass
        else:
            continue


        ###### sorting out PIaxes for MoI for UA fragment

        modPIx = PIaxes[0][0] ** 2 + PIaxes[0][1] ** 2 + PIaxes[0][2] ** 2
        modPIy = PIaxes[1][0] ** 2 + PIaxes[1][1] ** 2 + PIaxes[1][2] ** 2
        modPIz = PIaxes[2][0] ** 2 + PIaxes[2][1] ** 2 + PIaxes[2][2] ** 2


        PIaxes[0][0] = PIaxes[0][0] / (modPIx ** 0.5)
        PIaxes[0][1] = PIaxes[0][1] / (modPIx ** 0.5)
        PIaxes[0][2] = PIaxes[0][2] / (modPIx ** 0.5)

        PIaxes[1][0] = PIaxes[1][0] / (modPIy ** 0.5)
        PIaxes[1][1] = PIaxes[1][1] / (modPIy ** 0.5)
        PIaxes[1][2] = PIaxes[1][2] / (modPIy ** 0.5)

        PIaxes[2][0] = PIaxes[2][0] / (modPIz ** 0.5)
        PIaxes[2][1] = PIaxes[2][1] / (modPIz ** 0.5)
        PIaxes[2][2] = PIaxes[2][2] / (modPIz ** 0.5)



        #get dot product of Paxis1 and CoM->atom1 vect
        #will just be [0,0,0]
        RRaxis = vector(coord_list[0], COM, dimensions)
        #flip each Paxis if its pointing out of UA
        dotProd1 = np.dot(np.array(PIaxes[0]), RRaxis)
        if dotProd1 < 0:
            PIaxes[0][0] = -PIaxes[0][0]
            PIaxes[0][1] = -PIaxes[0][1]
            PIaxes[0][2] = -PIaxes[0][2]

        dotProd2 = np.dot(np.array(PIaxes[1]), RRaxis)
        if dotProd2 < 0:
            PIaxes[1][0] = -PIaxes[1][0]
            PIaxes[1][1] = -PIaxes[1][1]
            PIaxes[1][2] = -PIaxes[1][2]

        dotProd3 = np.dot(np.array(PIaxes[2]), RRaxis)
        if dotProd3 < 0:
            PIaxes[2][0] = -PIaxes[2][0]
            PIaxes[2][1] = -PIaxes[2][1]
            PIaxes[2][2] = -PIaxes[2][2]


    #MI = moment of inertia
    axis1MI = 0 #x
    axis2MI = 0 #y
    axis3MI = 0 #z
    #'''

    for coord, mass in zip(coord_list, mass_list):
        dx = coord[0] - COM[0]
        dy = coord[1] - COM[1]
        dz = coord[2] - COM[2]

        PIaxes_xx = PIaxes[0][1] * dz - PIaxes[0][2] * dy
        PIaxes_xy = PIaxes[0][2] * dx - PIaxes[0][0] * dz
        PIaxes_xz = PIaxes[0][0] * dy - PIaxes[0][1] * dx

        PIaxes_yx = PIaxes[1][1] * dz - PIaxes[1][2] * dy
        PIaxes_yy = PIaxes[1][2] * dx - PIaxes[1][0] * dz
        PIaxes_yz = PIaxes[1][0] * dy - PIaxes[1][1] * dx

        PIaxes_zx = PIaxes[2][1] * dz - PIaxes[2][2] * dy
        PIaxes_zy = PIaxes[2][2] * dx - PIaxes[2][0] * dz
        PIaxes_zz = PIaxes[2][0] * dy - PIaxes[2][1] * dx

        daxisx = (PIaxes_xx ** 2 + PIaxes_xy ** 2 + PIaxes_xz ** 2) ** 0.5
        daxisy = (PIaxes_yx ** 2 + PIaxes_yy ** 2 + PIaxes_yz ** 2) ** 0.5
        daxisz = (PIaxes_zx ** 2 + PIaxes_zy ** 2 + PIaxes_zz ** 2) ** 0.5

        axis1MI += (daxisx ** 2) * mass
        axis2MI += (daxisy ** 2) * mass
        axis3MI += (daxisz ** 2) * mass



    UA_moI = [axis1MI, axis2MI, axis3MI]

    return UA_moI, PIaxes



def principalAxesMOI(all_data, bonded_atoms_list, COM, Hs):
    '''
    Calculate the principal axes of the MOI matrix, then calc forces.
    If Hs is True, then consider H coords in MOI calcs.
    If Hs is False, then don't include H coords, instead add their mass
    onto UA heavy atom (HA).
    '''

    coord_list = []
    mass_list = []

    for atom in bonded_atoms_list:
        if Hs == True:
            #print (atom.atom_name)
            coord_list.append(atom.coords)
            mass_list.append(atom.mass)
        elif Hs == False:
            if atom.mass > 1.1:
                UA_mass = 0 #add mass of HA and bonded Hs
                #print (atom.atom_name)
                UA_mass += atom.mass
                coord_list.append(atom.coords) #only include coords of UA
                for h in atom.bonded_to_atom_num:
                    H = all_data[h]
                    if H.mass < 1.1:
                        UA_mass += H.mass
                    else:
                        continue
                mass_list.append(UA_mass) 
                        #include mass of Hs on heavy atom mass
        else:
            continue


    moI = MOI(COM, coord_list, mass_list)
    #if Hs == False:
        #print ('masses', mass_list)
        #print ('moi', moI)


    eigenvalues, eigenvectors = LA.eig(moI) 

    #different values generated to Jon's code
    #print (eigenvalues)
    #print (eigenvectors)
    transposed = np.transpose(eigenvectors) #turn columns to rows

    #bonded_atoms_list[0].WMprincipalAxis = transposed, eigenvalues, COM
    #print (moI, eigenvalues, eigenvectors)

    min_eigenvalue = abs(eigenvalues[0])
    if eigenvalues[1] < min_eigenvalue:
        min_eigenvalue = eigenvalues[1]
    if eigenvalues[2] < min_eigenvalue:
        min_eigenvalue = eigenvalues[2]

    max_eigenvalue = abs(eigenvalues[0])
    if eigenvalues[1] > max_eigenvalue:
        max_eigenvalue = eigenvalues[1]
    if eigenvalues[2] > max_eigenvalue:
        max_eigenvalue = eigenvalues[2]

    #print (min_eigenvalue * const, max_eigenvalue * const) 
    #same as Jons

    #'''
    #PA = principal axes
    axis1PA = [0, 0, 0] #[x,y,z]
    axis2PA = [0, 0, 0] #[x,y,z]
    axis3PA = [0, 0, 0] #[x,y,z]

    #MI = moment of inertia
    axis1MI = 0 #x
    axis2MI = 0 #y
    axis3MI = 0 #z
    #'''


    for i in range(0,3):
        if eigenvalues[i] == max_eigenvalue:
            axis1PA = transposed[i]
            axis1MI = eigenvalues[i]
        elif eigenvalues[i] == min_eigenvalue:
            axis3PA = transposed[i]
            axis3MI = eigenvalues[i]
        else:
            axis2PA = transposed[i]
            axis2MI = eigenvalues[i]




    principal_axes = [axis1PA, axis2PA, axis3PA]
    MI_axis = [axis1MI, axis2MI, axis3MI]

    #print ('mI', MI_axis, bonded_atoms_list[0].atom_num)

    return principal_axes, MI_axis



def rotateFT(all_data, bonded_atoms_list, WM_principal_axes, 
        UA_principal_axes, MI_axis, molecule_COM, UA_COM, Hs):
    '''
    '''

    ###lists used for torque calcs
    coord_list = []
    mass_list = []
    force_list = []
    forces_summed = np.zeros(3)

    for atom in bonded_atoms_list:
        if Hs == True:
            coord_list.append(atom.coords)
            mass_list.append(atom.mass)
            force_list.append(atom.forces)
            forces_summed += atom.forces
        elif Hs == False:
            if atom.mass > 1.1:
                #print (atom.atom_name)
                UA_mass = 0 #add mass of HA and bonded Hs
                UA_forces = np.zeros(3)
                UA_mass += atom.mass
                UA_forces += np.array(atom.forces)
                coord_list.append(atom.coords) #only include coords of UA
                forces_summed += atom.forces
                for h in atom.bonded_to_atom_num:
                    H = all_data[h]
                    if H.mass < 1.1:
                        UA_mass += H.mass
                        UA_forces += np.array(H.forces)
                        forces_summed += H.forces
                    else:
                        continue
                mass_list.append(UA_mass) #include mass of Hs on heavy atom mass
                force_list.append(UA_forces)
        else:
            continue

    torque = [0, 0, 0] #default is no torque
    ### calc torques here with COM and PA of molecule or UA
    if UA_principal_axes != None and MI_axis != None:
        if Hs == True:
            torque = calcTorque(UA_COM, coord_list, UA_principal_axes, 
                    force_list, MI_axis)
        if Hs == False:
            torque = calcTorque(molecule_COM, coord_list, WM_principal_axes, 
                    force_list, MI_axis)



    ### Rotate forces here
    F1 = np.dot(forces_summed, WM_principal_axes[0])
    F2 = np.dot(forces_summed, WM_principal_axes[1])
    F3 = np.dot(forces_summed, WM_principal_axes[2])
    mass_sqrt = sum(mass_list) ** 0.5
    force = np.array([float(F1)/float(mass_sqrt), 
            float(F2)/float(mass_sqrt), 
            float(F3)/float(mass_sqrt)])


    return force, torque




def calcTorque(cm, coord_list, principal_axes, force_list, MI_axis):
    '''
    '''
    x_cm, y_cm, z_cm = cm[0], cm[1], cm[2]
    #MI_axis = np.sqrt(MI_axis) #sqrt moi to weight torques

    MI_axis_sqrt = []
    #print (MI_axis)
    for coord in MI_axis:
        coord_round = round(coord, 10)
        if coord_round != 0:
            c_sqrt = coord ** 0.5
            MI_axis_sqrt.append(c_sqrt)
        elif coord_round == 0:
            MI_axis_sqrt.append(coord)
        else:
            continue

    MI_axis_sqrt = np.array(MI_axis_sqrt)
    #print (MI_axis, MI_axis_sqrt)

    W_torque = np.zeros(3)
    for coord, force in zip(coord_list, force_list):
        atom_torque = np.array([0, 0, 0])
        #count_atom += 1
        new_coords = []
        new_forces = []
        for axis in principal_axes:
            new_c = axis[0]*(coord[0] - x_cm) + axis[1]*(coord[1] - y_cm) \
                    + axis[2]*(coord[2] - z_cm)
            new_f = axis[0]*force[0] + axis[1]*force[1] + axis[2]*force[2]
            new_coords.append(new_c)
            new_forces.append(new_f)
        #print (new_coords)
        #print (force)
        #CROSS PRODUCT
        torquex = float(new_coords[1]*new_forces[2] \
                - new_coords[2]*new_forces[1]) #/ float(1e10)
        torquey = float(new_coords[2]*new_forces[0] \
                - new_coords[0]*new_forces[2]) #/ float(1e10)
        torquez = float(new_coords[0]*new_forces[1] \
                - new_coords[1]*new_forces[0]) #/ float(1e10)
        #print (torquex, torquey, torquez)

        atom_torque = (torquex, torquey, torquez)
        #atom_torque = np.divide(atom_torque, MI_axis_sqrt)

        atom_torque2 = []
        for t, MI in zip(atom_torque, MI_axis_sqrt):
            if MI != 0:
                t_divide = float(t) / float(MI)
                atom_torque2.append(t_divide)
            elif MI == 0:
                atom_torque2.append(0)
            else:
                continue

        atom_torque2 = np.array(atom_torque2)

        W_torque += atom_torque2

    return (W_torque)









import os, sys
import MDAnalysis as mda
from CodeEntropy.FunctionCollection import EntropyFunctions as EF
from CodeEntropy.ClassCollection import DataContainer as DC
import numpy as np
import pandas as pd
import pytest
from numpy import testing, array


@pytest.fixture(scope='session')
def data_dir():
    data_dir = os.path.dirname(os.path.abspath(__file__))
    return data_dir
@pytest.fixture(scope='session')
def dataContainer_DNA(data_dir):
    tprfile = os.path.join(data_dir,"data/md_A4_dna.tpr")
    trrfile = os.path.join(data_dir,"data/md_A4_dna_xf.trr")
    u = mda.Universe(tprfile, trrfile)
    dataContainer_DNA = DC.DataContainer(u)
    return dataContainer_DNA

@pytest.fixture(scope='session')
def dataContainer_protein(data_dir):
    tprfile = os.path.join(data_dir,"data/1AKI_ws.tpr")
    trrfile = os.path.join(data_dir,"data/1AKI_ws.trr")
    u = mda.Universe(tprfile, trrfile)
    dataContainer_protein = DC.DataContainer(u)
    return dataContainer_protein

@pytest.fixture(scope='session')
def tScale():
    tScale = 1.0
    return tScale
@pytest.fixture(scope='session')
def fScale():
    fScale = 1.0
    return fScale
@pytest.fixture(scope='session')
def temper():
    temper = 300.0 #K
    return temper
@pytest.fixture(scope='session')
def axis_list():
    axis_list = ["C5'", "C4'", "C3'"]
    return axis_list
@pytest.fixture(scope='session')
def thread():
    tScale = 2
    return tScale

ce_ref_index = [
    (75, 41),
    (66, 83), 
    (24, 17), 
    (88, 51), 
    (11, 82), 
]
@pytest.mark.parametrize("frame, atom, coords_ref",
    [
        (75, 41, array([25.98599243, 20.31201172, 16.64250374])),
        (66, 83, array([20.92900848, 19.77252197, 29.3363533 ])), 
        (24, 17, array([19.44181824, 24.54725456, 20.45416069])), 
        (88, 51, array([24.02828407, 25.08138275, 19.34505081])), 
        (11, 82, array([24.37817574, 25.81197739, 24.61750793])), 
    ]
)
def test_CodeEntropy_parser_labCoords(dataContainer_DNA, frame, atom, coords_ref):
    """
    sanity check for pulling coords from MDAnalysis.Universe
    """
    testing.assert_array_almost_equal(dataContainer_DNA._labCoords[frame][atom], coords_ref)

@pytest.mark.parametrize("frame, atom, coords_ref",
    [
        (75, 41, array([ -0.62618136,   4.76334143, -11.99212551])),
        (66, 83, array([ 39.02585983, 110.13354492,   5.35777664])), 
        (24, 17, array([  5.45531702,  20.02843094, -18.68554306])), 
        (88, 51, array([  50.23020554,   30.55683327, -121.57384491])), 
        (11, 82, array([-11.61987019,   8.08342552,  21.33005142])), 
    ]
)
def test_CodeEntropy_parser_labForces(dataContainer_DNA, frame, atom, coords_ref):
    """
    sanity check for pulling forces from MDAnalysis.Universe
    """
    testing.assert_array_almost_equal(dataContainer_DNA._labForces[frame][atom], coords_ref)

def test_CodeEntropy_parser_isBBAtomArray(dataContainer_protein):
    """
    sanity check for pulling atom selection boolean from MDAnalysis.Universe.select_atom
    """
    ref = array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 0., 1., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 0., 1.,
       0., 0., 0., 0., 0., 1., 1., 1., 0., 1., 0., 0., 0., 0., 0., 1., 1.,
       1., 0., 1., 0., 0., 0., 0., 0., 1., 1., 1., 0., 1., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 0., 1., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 0.,
       1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 1., 1., 1., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 1., 1., 1., 0., 1., 0., 0., 1., 1., 1., 0., 1., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 0., 1.,
       0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 0., 1., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 1., 1., 1., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 1., 1., 1., 0.])
    testing.assert_array_almost_equal(dataContainer_protein.isBBAtomArray[104:320], ref)

def test_CodeEntropy_parser_isCAtomArray(dataContainer_protein):
    """
    sanity check for pulling atom selection boolean from MDAnalysis.Universe.select_atom
    """
    ref = array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0.])
    testing.assert_array_almost_equal(dataContainer_protein.isCAtomArray[104:320], ref)

def test_CodeEntropy_parser_isCaAtomArray(dataContainer_protein):
    """
    sanity check for pulling atom selection boolean from MDAnalysis.Universe.select_atom
    """
    ref = array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 1., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.])
    testing.assert_array_almost_equal(dataContainer_protein.isCaAtomArray[104:320], ref)

def test_CodeEntropy_parser_isHydrogenArray(dataContainer_protein):
    """
    sanity check for pulling atom selection boolean from MDAnalysis.Universe.select_atom
    """
    ref = array([1., 0., 1., 1., 0., 1., 1., 0., 0., 0., 0., 0., 0., 1., 0., 1., 0.,
       1., 1., 0., 1., 0., 1., 1., 1., 0., 1., 1., 1., 0., 0., 0., 1., 0.,
       1., 0., 1., 1., 1., 0., 0., 0., 1., 0., 1., 0., 1., 1., 1., 0., 0.,
       0., 1., 0., 1., 0., 1., 1., 1., 0., 0., 0., 1., 0., 1., 0., 1., 1.,
       0., 1., 1., 0., 0., 1., 1., 1., 0., 0., 0., 1., 0., 1., 0., 1., 1.,
       0., 1., 1., 0., 1., 1., 0., 1., 1., 0., 1., 1., 1., 0., 0., 0., 1.,
       0., 1., 0., 1., 1., 0., 1., 1., 0., 1., 1., 0., 1., 0., 0., 1., 1.,
       0., 1., 1., 0., 0., 0., 1., 0., 1., 0., 1., 1., 0., 0., 0., 1., 0.,
       1., 0., 1., 0., 0., 0., 1., 0., 1., 1., 0., 0., 0., 1., 0., 1., 0.,
       1., 1., 0., 1., 0., 1., 1., 1., 0., 1., 1., 1., 0., 0., 0., 1., 0.,
       1., 0., 1., 1., 0., 0., 0., 0., 0., 0., 1., 0., 1., 0., 1., 1., 0.,
       0., 0., 1., 1., 0., 0., 0., 1., 0., 1., 0., 1., 1., 0., 0., 1., 0.,
       1., 0., 1., 0., 1., 0., 0., 1., 0., 0., 0., 1.])
    testing.assert_array_almost_equal(dataContainer_protein.isHydrogenArray[104:320], ref)


def test_CodeEntropy_parser_isNAtomArray(dataContainer_protein):
    """
    sanity check for pulling atom selection boolean from MDAnalysis.Universe.select_atom
    """
    ref = array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0.])
    testing.assert_array_almost_equal(dataContainer_protein.isNAtomArray[104:320], ref)

def test_CodeEntropy_parser_isOAtomArray(dataContainer_protein):
    """
    sanity check for pulling atom selection boolean from MDAnalysis.Universe.select_atom
    """
    ref = array([0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 1., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0., 0.,
       0., 0., 0., 0., 0., 0., 0., 0., 0., 1., 0., 0.])
    testing.assert_array_almost_equal(dataContainer_protein.isOAtomArray[104:320], ref)

def test_CodeEntropy_whole_molecule(dataContainer_DNA, tScale, fScale, temper):
    """test for computing entropy at whole molecule level"""
    wm_entropyFF, wm_entropyTT = EF.compute_entropy_whole_molecule_level(
        arg_hostDataContainer = dataContainer_DNA,
        arg_outFile = None,
        arg_selector = "all", 
        arg_moutFile = None,
        arg_nmdFile = None,
        arg_fScale = fScale,
        arg_tScale = tScale,
        arg_temper = temper,
        arg_verbose = 5
    )
    assert wm_entropyFF == pytest.approx(47.230063169278864)
    assert wm_entropyTT == pytest.approx(44.35619230374928)

def test_CodeEntropy_res_level(dataContainer_DNA, tScale, fScale, temper, axis_list):
    """test for computing entropy at residue level"""
    res_entropyFF, res_entropyTT = EF.compute_entropy_residue_level(
        arg_hostDataContainer = dataContainer_DNA,
        arg_outFile = None,
        arg_selector = 'all', 
        arg_moutFile = None,
        arg_nmdFile = None,
        arg_fScale = fScale,
        arg_tScale = tScale,
        arg_temper = temper,
        arg_verbose = 5,
        arg_axis_list = axis_list,
    )
    assert res_entropyFF == pytest.approx(194.93511803752938)
    assert res_entropyTT == pytest.approx(285.7213867860228)


def test_CodeEntropy_united_atom_level(dataContainer_DNA, tScale, fScale, temper, axis_list, data_dir):
    """test for computing entropy at united atom level"""
    UA_entropyFF, UA_entropyTT, res_df = EF.compute_entropy_UA_level(
        arg_hostDataContainer = dataContainer_DNA,
        arg_outFile = None,
        arg_selector = 'all', 
        arg_moutFile = None,
        arg_nmdFile = None,
        arg_fScale = fScale,
        arg_tScale = tScale,
        arg_temper = temper,
        arg_verbose = 1,
        arg_axis_list = axis_list,
        arg_csv_out= None,
    )
    
    assert UA_entropyFF == pytest.approx(1439.501746003283)
    assert UA_entropyTT == pytest.approx(107.67610435497281)
    csv_dir = os.path.join(data_dir,"data/Atom_level_res_data.csv")
    res_ref = pd.read_csv(csv_dir, na_values="nan")
    pd.testing.assert_frame_equal(res_ref, res_df, check_dtype=False)

def test_CodeEntropy_united_atom_level_multiprocess(dataContainer_DNA, tScale, fScale, temper, axis_list, thread, data_dir):
    """test for computing entropy at united atom level with multiprocess"""
    UA_entropyFF, UA_entropyTT, res_df = EF.compute_entropy_UA_level_multiprocess(
        arg_hostDataContainer = dataContainer_DNA,
        arg_outFile = None,
        arg_selector = 'all', 
        arg_moutFile = None,
        arg_nmdFile = None,
        arg_fScale = fScale,
        arg_tScale = tScale,
        arg_temper = temper,
        arg_verbose = 1,
        arg_csv_out= None,
        arg_axis_list = axis_list,
        arg_thread= thread,
    )
    assert UA_entropyFF == pytest.approx(1439.5017460032827)
    assert UA_entropyTT == pytest.approx(107.67610435497281)
    csv_dir = os.path.join(data_dir,"data/Atom_level_res_data.csv")
    res_ref = pd.read_csv(csv_dir, na_values="nan")
    pd.testing.assert_frame_equal(res_ref, res_df, check_dtype=False)

def test_CodeEntropy_topo_method5(dataContainer_protein):
    """test for computing topographical entroy using method 5 AEM method. AEM is random so it is impossible to make test accurately"""
    result_entropyAEM = EF.compute_topographical_entropy_AEM(
        arg_hostDataContainer = dataContainer_protein, 
        arg_selector = "all",
        arg_outFile = None, 
        arg_verbose = 5
    )
    assert 220.0 < result_entropyAEM < 230.0

# def test_CodeEntropy_topo_method3(dataContainer_DNA):
#     """test for computing topographical entroy using method 3 Correlation density method"""
#     result_entropy3 = EF.compute_topographical_entropy_method3(
#         arg_hostDataContainer = dataContainer_DNA, 
#         arg_selector = "all",
#         arg_outFile = None, 
#         arg_verbose = 5
#     ) 
#     result = np.sum(result_entropy3)
#     reference = 211.99999999999983
#     assert reference == pytest.approx(result)

def test_CodeEntropy_topo_method0_SC(dataContainer_protein):
    """test for computing topographical entroy using method 3 Correlation density method"""
    result_entropy0_SC = EF.compute_topographical_entropy0_SC(
        arg_hostDataContainer = dataContainer_protein, 
        arg_selector = "all",
        arg_outFile = None, 
        arg_verbose = 5
    )

    reference = 621.2072742842519
    assert reference == pytest.approx(result_entropy0_SC)

def test_CodeEntropy_topo_method0_BB(dataContainer_protein):
    """test for computing topographical entroy using method 3 Correlation density method"""
    result_entropy0_BB = EF.compute_topographical_entropy0_BB(
        arg_hostDataContainer = dataContainer_protein, 
        arg_selector = "all",
        arg_outFile = None, 
        arg_verbose = 5
    )

    reference = 728.8710286898643
    assert reference == pytest.approx(result_entropy0_BB)

def test_CodeEntropy_topo_method1_SC(dataContainer_protein):
    """test for computing topographical entroy using method 1 some degree of randomness so a range is used"""
    result_entropy1_SC = EF.compute_topographical_entropy1_SC(
        arg_hostDataContainer = dataContainer_protein, 
        arg_selector = "all",
        arg_outFile = None, 
        arg_verbose = 5
    )

    reference = 390.587
    assert 388.0 <= result_entropy1_SC <= 395.0

def test_CodeEntropy_topo_method1_BB(dataContainer_protein):
    """test for computing topographical entroy using method 1"""
    result_entropy1_BB = EF.compute_topographical_entropy1_BB(
        arg_hostDataContainer = dataContainer_protein, 
        arg_selector = "all",
        arg_outFile = None, 
        arg_verbose = 5
    )

    reference = 38.37223423319515
    assert reference == pytest.approx(result_entropy1_BB)



""" A module/collection of classes that store information 
about the bonded properties of a molecule. E.g. bonds, angles, dihedrals"""

import numpy as nmp
import sys

from CodeEntropy.FunctionCollection import GeometricFunctions as GF
from CodeEntropy.FunctionCollection import CustomFunctions as CF
from CodeEntropy.ClassCollection import CustomDataTypes as CDT
from CodeEntropy.FunctionCollection import Utils

class BondedStructure(object):
	"""
	A parent class designed to store information about 
	bonded atoms in a structure. Bonded atoms can form
	a bond, an angle, a dihedral and an improper.
	"""

	def __init__(self, atomIndices, arg_container):
		self.atomList = atomIndices
		self.container = arg_container
		self.qvalue = None

	def __hash__(self):
		return hash(('{}_'*len(self.atomList)).format(*sorted(self.atomList)))

	def set_quantity(self, value):
		"""Set the value for the relevant geomtric quantity associated
		with the bonded structure. For bonds - distance, angles - angle, 
		dihedral - angle, improper - angle.
		"""
		self.qvalue = value

	def get_quantity(self):
		"""Get the value for the relevant geomtric quantity associated
		with the bonded structure. For bonds - distance, angles - angle, 
		dihedral - angle, improper - angle.
		"""
		return self.qvalue

	def is_heavy(self):
		"""Return true if all of the atoms are not
		hydrogens"""
		retVal = True 
		for iAtom in self.atomList:
			if self.container.isHydrogenArray[iAtom]:
				retVal = False
				break
		return retVal
#END

class Bond(BondedStructure):
	""" Bond between two atoms (preferably covalent). 
	May serve as a parent of 'contact' """

	def __init__(self, arg_atomIndices, arg_container):
		super().__init__(arg_atomIndices, arg_container)
		self.atom1, self.atom2 = arg_atomIndices # indices of 2 atoms

#END

class Angle(BondedStructure):
	""" Angle of three atoms. """
	def __init__(self, arg_atomIndices, arg_container):
		super().__init__(arg_atomIndices, arg_container)
#END

class Dihedral(BondedStructure):
	""" Dihedral composed of 4 positions (2 planes with an intersecting line) """

	def __init__(self, arg_atomIndices, arg_container):
		super().__init__(arg_atomIndices, arg_container)
		self.atom1, self.atom2, self.atom3, self.atom4 = arg_atomIndices	 # indices of the 4 atoms

	def __str__(self):
		# return 'Dihedral: {} -> {} -> {} -> {}'.format(self.atom1, self.atom2, self.atom3, self.atom4)
		return '{} {} {} {} '.format(*self.atomList)


	def __lt__(self, arg_other):
		return sorted(self.atomList[1:3]) < sorted(arg_other.atomList[1:3])

	def __gt__(self, arg_other):
		return sorted(self.atomList[1:3]) > sorted(arg_other.atomList[1:3])

	# def __eq__(self, arg_other):
	#	 """ Two dihedrals are deemed equal if their 2,3 atoms are the same """
	#	 return sorted(self.atomList[1:3]) == sorted(arg_other.atomList[1:3])
	

	def is_BB_dihedral(self):
		""" Based on the information in the base molecule's arrays, see if 2-3 atoms 
		in this dihedral are BB atoms """
		retBool = self.container.isBBAtomArray[self.atom2] 
		retBool = retBool and self.container.isBBAtomArray[self.atom3] 
		return retBool
	#END

	def is_BB_psi(self):
		""" Checks if a heavy dihedral is BB psi (X-C-----Ca-Y)"""
		if self.is_BB_dihedral():
			if self.container.isCAtomArray[self.atom2] and self.container.isCaAtomArray[self.atom3]:
				return True
			elif self.container.isCAtomArray[self.atom3] and self.container.isCaAtomArray[self.atom2]:
				return True
			else:
				return False
		else:
			# if not a BB dihedral, it aint a PHI or PSI
			return False
	#END

	def is_BB_phi(self):
		""" Checks if a heavy dihedral is BB phi (X-Ca-----N-Y)"""
		if self.is_BB_dihedral():
			if self.container.isNAtomArray[self.atom2] and self.container.isCaAtomArray[self.atom3]:
				return True
			elif self.container.isNAtomArray[self.atom3] and self.container.isCaAtomArray[self.atom2]:
				return True
			else:
				return False
		else:
			# if not a BB dihedral, it aint a PHI or PSI
			return False
	#END

	def is_from_same_residue(self):
		""" Based on the information in the base molecule's arrays, see if the two central atoms (2,3)
		in this dihedral belong to the same residue. Return the residue index if yes, else -100"""
		if self.container.universe.atoms.resids[self.atom2] ==\
		self.container.universe.atoms.resids[self.atom3]:
			return self.container.universe.atoms.resids[self.atom2]
		else:
			return -10000
	#END

	def get_dihedral_angle_lab(self, arg_frame):
		""" Using the function compute_dihedral in GeometricFunctions module, return the dihedral for this quadruplet in the lab frame"""
		r1 = self.container._labCoords[arg_frame][self.atom1]
		r2 = self.container._labCoords[arg_frame][self.atom2]
		r3 = self.container._labCoords[arg_frame][self.atom3]
		r4 = self.container._labCoords[arg_frame][self.atom4]

		return GF.compute_dihedral(r1, r2, r3, r4)
	#END

	def get_dihedral_angle_local(self, arg_frame):
		""" Using the function compute_dihedral in GeometricFunctions module, return the dihedral for this quadruplet in the local frame"""
		r1 = self.container.localCoords[arg_frame][self.atom1]
		r2 = self.container.localCoords[arg_frame][self.atom2]
		r3 = self.container.localCoords[arg_frame][self.atom3]
		r4 = self.container.localCoords[arg_frame][self.atom4]

		return GF.compute_dihedral(r1, r2, r3, r4)
	#END

	def get_state_ts(self, arg_verbose = 3, arg_bw = 30):
		"""
		Create a state vector, showing the state in which the input dihedral is
		as a function of time.
		The function creates a histogram from the timeseries of the dihedral angle values
		and identifies points of dominant occupancy (called CONVEX TURNING POINTS).
		Based on the identified TPs, states are assigned to each configuration of the dihedral.
		
		Return
		------
		A timeseries with integer labels describing the state at each point in time.
		
		"""
		cosBW = nmp.cos(nmp.deg2rad(arg_bw))
		numFrames = len(self.container.trajSnapshots)
		phiTS = nmp.zeros(numFrames)
		stateTS = nmp.zeros(numFrames)
		
		for iFrame in range(numFrames):

			# fetch the dihedral value at that frame
			phi = self.get_dihedral_angle_lab(arg_frame = iFrame)
			if phi < 0:
				phi += 360
				
			phiTS[iFrame] = phi
			
		# create a histogram using numpy
		popul, binEdges = nmp.histogram(a = phiTS, \
										bins = nmp.linspace(start=0,stop=360, num = 1 + int(360/arg_bw), endpoint=True))
		binVal = [0.5 * (binEdges[i] + binEdges[i+1]) for i in range(0,len(popul))]

		# create a circular doubly linked list of the bin elements.
		# CCDL data structure will allow to iterate through the list
		# without worrying about histogram's end-points
		histCDLL = CDT.CircularDoublyLinkedList()
		for ele in zip(binVal, popul):
			newNode = CDT.DoublyLinkedListNode(arg_value=ele)
			histCDLL.append(newNode)
		
		# identify "convex turning-points" and populate a list of TPs
		# TP := a bin whose neighboring bins have smaller population
		tpList = []
		for iNode in histCDLL.iterator_cw():
			prevBin, prevPopul = iNode.previousNode.value
			currBin, currPopul = iNode.value
			nextBin, nextPopul = iNode.nextNode.value

			if currPopul >= prevPopul and currPopul >= nextPopul:
				# the equality condition is greedy
				# it is expected to put empty bins with 
				# empty neighbors in the list.
				# They will be filtered out in the next step

				# Another condition is that TPs should be at least arg_bw (default: 30 deg) apart
				# Keepin gin mind that angle is cicularly preiodic but the histogram's x-axis
				# is linear, the difference is measured by comparing the dot product of 
				# e^i(angle1) and e^i(agnle2) with cosine(arg_bw). Isnt that COOL?

				newTP = iNode.value
				newAngle = newTP[0]
				newVec = CF.euler_vector(arg_r = 1., arg_theta = newAngle)
				goodTP = True
				for tp in tpList:
					tpAngle = tp[0]
					tpVec = CF.euler_vector(arg_r = 1., arg_theta = tpAngle)
					
					dotProd = CF.dot_product(newVec, tpVec)
					if nmp.isclose(dotProd, cosBW) or dotProd > cosBW:
						goodTP = False

				if goodTP:
					tpList.append(newTP)
					
		# filter out the empty turning points
		tpList = list(filter(lambda tp: tp[1] != 0, tpList))

		# TP angles
		tpVals = [tp[0] for tp in tpList]
		
		if arg_verbose > 3:
			Utils.printflush('{} turning point(s) found for dihedral {}.'.format(len(tpVals), self.atomList), end=' ')
		
		if arg_verbose >= 5:
			Utils.printflush(':', end=' ')
			for v in tpVals:
				Utils.printflush(v, end=' ')
			Utils.printflush('')

		# go through each frame again and assign state
		for iFrame, iPhi in enumerate(phiTS):
			# find the TP that the snapshot is least distant from
			diffList = [abs(iPhi - tp) for tp in tpVals]
			stateTS[iFrame] = nmp.argmin(diffList)
		
		return stateTS


import numpy as nmp
import sys

from bitarray import bitarray
from CodeEntropy import ClassCollection as CC 
from CodeEntropy.FunctionCollection import Utils

def cross_product(arg_v1, arg_v2):
	""" 
	Returns cross product of two vectors using analytical expressions 
	that one would have from using the determinant used to compute the product.
	It appears to be a tad bit faster than the numpy one.
	"""
	vecCross = nmp.zeros(3)

	vecCross[0] =  arg_v1[1]*arg_v2[2] - arg_v2[1]*arg_v1[2]
	vecCross[1] = -arg_v1[0]*arg_v2[2] + arg_v2[0]*arg_v1[2]
	vecCross[2] =  arg_v1[0]*arg_v2[1] - arg_v2[0]*arg_v1[1]
		 
	return vecCross

#END

def dot_product(arg_v1, arg_v2):
	"""
	Returns dot product of two vectors using numpy's dot function.
	""" 
	return nmp.dot(arg_v1, arg_v2)
#END

def euler_vector(arg_r, arg_theta):
	"""
	Create and return a vector of the form r*e^i(theta) in the argand plane.
	Angle should be degrees.
	"""
	vx = arg_r * nmp.cos(nmp.deg2rad(arg_theta))
	vy = arg_r * nmp.sin(nmp.deg2rad(arg_theta))

	return (vx, vy)
#END


def filter_zero_rows_columns(arg_matrix):
	
	#record the initial size
	initShape = nmp.shape(arg_matrix)

	zeroIdx = list(filter(lambda row : nmp.all(nmp.isclose(arg_matrix[row,:] , 0.0)) , nmp.arange(nmp.shape(arg_matrix)[0])))
	allIdx = nmp.ones((nmp.shape(arg_matrix)[0]), dtype=bool)
	allIdx[zeroIdx] = False 
	arg_matrix = arg_matrix[allIdx,:]

	allIdx = nmp.ones((nmp.shape(arg_matrix)[1]), dtype=bool)
	zeroIdx = list(filter(lambda col : nmp.all(nmp.isclose(arg_matrix[:,col] , 0.0)) , nmp.arange(nmp.shape(arg_matrix)[1])))
	allIdx[zeroIdx] = False
	arg_matrix = arg_matrix[:,allIdx]

	# get the final shape
	finalShape = nmp.shape(arg_matrix)

	if initShape != finalShape:
		Utils.printflush('A shape change has occured ({},{}) -> ({}, {})'.format(*initShape, *finalShape))

	return arg_matrix
#END


def probability_of_coexistence(arg_v1, arg_v2):
	""" From the input vectors/list , the function returns the probability
	 of co-occurence (True value for both vectors at the same index/time) of the same state across them """
	
	try:
		assert(len(arg_v1) == len(arg_v2))
	except:
		raise ValueError("Unequal vector lengths ({} and {}). The lengths should be the same.".format(len(arg_v1), len(arg_v2)))
	
	#prob   = nmp.dot(arg_v1, arg_v2)/len(arg_v1)
	prob = 0
	for x, y in zip(arg_v1, arg_v2):
		prob += (x * y)

	prob /= len(arg_v1)

	if prob < 0:
		Utils.printflush(arg_v1)
		Utils.printflush(arg_v2)

	return prob
#END

def covariance(arg_v1, arg_v2):
	""" Returns the covariance between two vectors/list """

	mean1 = nmp.mean(arg_v1)
	mean2 = nmp.mean(arg_v2)
	
	try:
		assert(len(arg_v1) == len(arg_v2))
	except:
		raise ValueError("Inconsistent vector lengths ({} and {}). The lengths should be the same.".format(len(arg_v1), len(arg_v2)))
		
	N = len(arg_v1)
	cov = 0
	for x, y in zip():
		cov += ((x - mean1) * (y - mean2))
	
	cov /= (N -  1)
	return cov

#END

def phi_coeff(arg_v1, arg_v2):
	"""
	From the input binary vector/list, creates a contingency matrix with the following entries:
	[1,1] <- probability that the order combination is 1/1 at the same time
	[1,2] <- probability that the order combination is 1/0 at the same time
	[2,1] <- probability that the order combination is 0/1 at the same time
	[2,2] <- probability that the order combination is 1/1 at the same time

	Computes and returns the phi coefficient from this matrix.
	"""
	contingencyMatrix = nmp.zeros((2,2))

	notV1 = nmp.array([not(bit) for bit in arg_v1], dtype = nmp.int8)
	notV2 = nmp.array([not(bit) for bit in arg_v2], dtype = nmp.int8)

	contingencyMatrix[0,0] = probability_of_coexistence(arg_v1, arg_v2)
	contingencyMatrix[0,1] = probability_of_coexistence(arg_v1, notV2)
	contingencyMatrix[1,0] = probability_of_coexistence(notV1, arg_v2)
	contingencyMatrix[1,1] = probability_of_coexistence(notV1, notV2)

	# print contingency matrix
	Utils.printflush('{:>8.6f} {:>8.6f} \n {:>8.6f} {:>8.6f}'. format(*nmp.ndarray.flatten(contingencyMatrix)))

	# column and row sums
	
	#   row sums
	sumR0 = nmp.sum(contingencyMatrix[0,:])
	sumR1 = nmp.sum(contingencyMatrix[1,:]) 

	# col sums
	sumC0 = nmp.sum(contingencyMatrix[:,0])
	sumC1 = nmp.sum(contingencyMatrix[:,1])
	
	Utils.printflush('r0 = {:>8.6f}, r1 = {:>8.6f}, c0 = {:>8.6f}, c1 = {:>8.6f}'.format(sumR0, sumR1, sumC0, sumC1))

	# return coeff
	if ( sumR0 * sumR1 * sumC0 * sumC1 == 0 ):
		return 0
	else:
		phi = nmp.linalg.det(contingencyMatrix)
		phi /= nmp.sqrt(sumR0 * sumR1 * sumC0 * sumC1)
		return phi
#END





# TEMPERATURE
DEF_TEMPER = 298 #K

# PHYSICAL CONSTANTS
N_AVOGADRO = 6.0221415e+23
GAS_CONST= 8.3144598484848  # J/mol/K
PLANCK_CONST = 6.62607004081818e-34  #J s
C_LIGHT = 29979245800 #cm/s

# MACHINE DATA
LITTLE_ENN="<"
BIG_ENN=">"

# LENGTH
NM2ANG = 1e1
M2ANG = 1e10
ANG2M = 1e-10
NM2M = 1e-9
ANG2NM = 1e-1


# MASS
AMU2KG = 1.66053892110e-27
KG2AMU = 6.022141289754078e+26
sqrtKG2AMU = 24540051527562.199

# ENERGY
KJ2KCAL = 0.239006
KCAL2KJ = 4.184

#CHARMM units
AKMA2PS = 4.88882129e-02  

def change_lambda_units(arg_lambdas):
	"""Unit of lambdas : kJ2 mol-2 A-2 amu-1
	change units of lambda to J/s2"""
	# return arg_lambdas * N_AVOGADRO * N_AVOGADRO * AMU2KG * 1e-26
	return arg_lambdas * 1e+29 / N_AVOGADRO

def get_KT2J(arg_temper):
	"""A temperature dependent KT to Joule conversion"""
	return 4.11e-21 * arg_temper/DEF_TEMPER





#!/usr/bin/env python

import sys
import math
import numpy as np
import logging

import MDAnalysis
from MDAnalysis import *
from MDAnalysis.analysis import distances

from CodeEntropy.poseidon.extractData.generalFunctions import distance, calcWaterSoluteAngle

from collections import defaultdict
from collections import Counter

from datetime import datetime

nested_dict = lambda: defaultdict(nested_dict) 







def getShellAssignment(all_data, excludedResnames, dimensions, 
        startTime, verbosePrint):
    '''
    Assign solvent to solvation shells
    '''


    nearestNonlike_dict = {} #need a dictionary of all UAs that are 
        #assigned as nearest non-likes with values as a list of UAs

    if excludedResnames == None:
        excludedResnames = []

    for x in range(0, len(all_data)):
        atom = all_data[x]
        if atom.mass > 1.1 and atom.nearest_sorted_array != None:
            #and len(atom.RAD) != 0:
            nearestNonlike = None
            nearestNonlikeDist = None
            for atomDist in atom.nearest_sorted_array:
                if nearestNonlikeDist == None:
                    nonlike = all_data[atomDist[0]]
                    if nonlike.resid != atom.resid \
                            and nonlike.atom_num \
                            not in atom.bonded_to_atom_num \
                            and nonlike.atom_num != atom.atom_num \
                            and nonlike.resname != atom.resname \
                            and nonlike.resname not in \
                                    excludedResnames \
                            and len(nonlike.RAD) != 0:
                        nearestNonlike = nonlike
                        nearestNonlikeDist = atomDist[1]
                else:
                    break

            #### When nearest nonlike solute is found
            if nearestNonlikeDist != None:
                atom.nearestAnyNonlike = (nearestNonlike, 
                        nearestNonlikeDist)

                if nearestNonlike.resid not in nearestNonlike_dict:
                    nearestNonlike_dict[nearestNonlike.resid] = [[], []]
                if nearestNonlike not in \
                        nearestNonlike_dict[nearestNonlike.resid][0]:
                    nearestNonlike_dict[nearestNonlike.resid][0].append(
                            nearestNonlike)
                if atom not in nearestNonlike_dict[nearestNonlike.resid][1]:
                    nearestNonlike_dict[nearestNonlike.resid][1].append(atom)
            else:
                continue
        else:
            continue


    verbosePrint('NEAREST NON-LIKE ASSIGNMENT')
    verbosePrint(datetime.now() - startTime)
    sys.stdout.flush()




    #### Part 2: For each nonlike UA with assigned molecules, get prox shell
    #### This is the slow part of the code - prox shell assignment

    #### Find closest nonlike for each molecule, needed below
    #### so only one atom per molecule is used in prox assignment, rest
    #### of atoms in molecules are given the same prox as nearest atom
    molecule_resid_nearest_dict = {}
    for x in range(0, len(all_data)):
        atom = all_data[x]
        if atom.nearestAnyNonlike != None:
            nearestNonlike = atom.nearestAnyNonlike[0]
            dist = atom.nearestAnyNonlike[1]
            if atom.resid not in molecule_resid_nearest_dict:
                molecule_resid_nearest_dict[atom.resid] = \
                        [nearestNonlike.resid, dist]
            elif atom.resid in molecule_resid_nearest_dict:
                if dist < molecule_resid_nearest_dict[atom.resid][1]:
                    molecule_resid_nearest_dict[atom.resid] = \
                            [nearestNonlike.resid, dist]
                else:
                    continue
            else:
                continue


    multiple_shellList = []
    multiple_atomNumList = []
    localMols = 0

    for resid, nonlike_atom_lists in nearestNonlike_dict.items():
        nonlike_list = nonlike_atom_lists[0]
        atom_list = nonlike_atom_lists[1]

        for nonlike in nonlike_list:

            if len(nonlike.RAD) != 0:
                for neighbour in nonlike.RAD:
                    X = neighbour
                    if X in atom_list and X.atom_num not in \
                            multiple_atomNumList and \
                            molecule_resid_nearest_dict[X.resid][0] == resid:
                        for molecule_atom in X.molecule_atomNums:
                            ma = all_data[molecule_atom] 
                            multiple_shellList.append([ma, 1])
                            multiple_atomNumList.append(ma.atom_num)
                            ma.hydrationShellRAD_solute = 1
                            localMols += 1
                    else:
                        continue

            if len(nonlike.RAD) == 0:
                logging.warning('no RAD nonlike:', nonlike)



    ##Anything beyond the 1st shell is assigned to 2nd
    for x in range(0, len(all_data)):
        atom = all_data[x]
        if atom.hydrationShellRAD_solute == None and atom.mass > 1.1:
            atom.hydrationShellRAD_solute = 2
        else:
            continue








def moleculePositionRankingRAD(all_data, waterTuple, dimensions):
    '''
    For each UA, rank its RAD shell by proximity to that
    references UAs nearest Non-like molecule UA.
    num = RAD shell from same molecule type, when nearest nonlike 
        resid is the same as the reference.
    'X' = when same molecule type has different nearest nonlike resid.
    'RESNAME' = when molecule of different type is in RAD shell.
    '0' = closest different type molecule in RAD shell. 
            (the one its assigned to, its nearest non-like!)


    Extra:
    for each 'X' found in a RAD shell, use this information
    to find neighbouring zones. Think voronoi but with zones.
    For each molecule that defines a zone, find it's neighbouring zones
    from X's in shell type.


    The following should not occur with new HB defn (HB in RAD shell only):
    'AX' = acceptors not in RAD shell but same molecule type as ref.
    'AS' = acceptors not in RAD shell and diff mol. type to ref.
    'DX' = donor equivalent to 'AX'
    'DS' = donor equivalent to 'AS'
    '''

    for x in range(0, len(all_data)):
        atom = all_data[x]
        if atom.mass > 1.1:

            #### part 1: get ranking of RAD Nc according to extended shells.
            shell_list_ranked = []
            shell_list_atom_nums = []

            for neighbour in atom.RAD:
                shell_list_atom_nums.append(neighbour.atom_num)
                ##include Hs in shell_list_atom_nums
                for bonded in neighbour.bonded_to_atom_num:
                    b = all_data[bonded]
                    if b.mass < 1.1:
                        shell_list_atom_nums.append(b.atom_num)
                    else:
                        continue

            if len(atom.RAD) != 0 and \
                    atom.nearestAnyNonlike != None:

                for neighbour in atom.RAD:
                    if neighbour.nearestAnyNonlike != None:
                        if neighbour.nearestAnyNonlike[0].resid == \
                                atom.nearestAnyNonlike[0].resid \
                                and neighbour.resname == atom.resname:
                            shell_list_ranked.append(
                                (str(neighbour.hydrationShellRAD_solute),
                                    neighbour.atom_num))

                        if neighbour.nearestAnyNonlike[0].resid != \
                                atom.nearestAnyNonlike[0].resid \
                                and neighbour.resname == atom.resname:
                            shell_list_ranked.append(('X', 
                                neighbour.atom_num))


                        if neighbour.resname != atom.resname \
                                and neighbour.resid == \
                                atom.nearestAnyNonlike[0].resid:
                            if neighbour.atom_num == \
                                    atom.nearestAnyNonlike[0].atom_num:
                                shell_list_ranked.append(('0',
                                    neighbour.atom_num))
                            else:
                                shell_list_ranked.append(('%s_%s' % (
                                    neighbour.resname, neighbour.atom_name),
                                    neighbour.atom_num))


                        if neighbour.resid != \
                                atom.nearestAnyNonlike[0].resid \
                                and neighbour.resname != atom.resname:
                            shell_list_ranked.append(('%s_%s' % (
                                neighbour.resname, neighbour.atom_name),
                                neighbour.atom_num))


                    else:
                        continue


            #### part 2: get HBing info wrt RAD Nc ranking, only for water

            acceptor_list = []
            donor_list = []

            if atom.hydrationShellRAD_solute == 1 and \
                    atom.resname in waterTuple:

                #what Hs is atom accepting from:
                if len(atom.nearest_Hs) != 0:
                    for H in atom.nearest_Hs:
                        for bonded_atom_num in H.bonded_to_atom_num:
                            bonded = all_data[bonded_atom_num]
                            if bonded.mass > 1.1:
                                X = bonded
                                if X.atom_num in shell_list_atom_nums:
                                    for rank_num in shell_list_ranked:
                                        rank = rank_num[0]
                                        num = rank_num[1]
                                        if num == X.atom_num:
                                            acceptor_list.append((rank, num))
                                        else:
                                            continue
                                elif X.atom_num not in shell_list_atom_nums \
                                        and X.resname == atom.resname:
                                            acceptor_list.append(('AX', 
                                                X.atom_num))
                                            logging.warning('AX')

                                elif X.atom_num not in shell_list_atom_nums \
                                        and X.resname != atom.resname:
                                            acceptor_list.append(('AS', 
                                                X.atom_num))
                                            logging.warning('AS')
                                else:
                                    continue



                #what atoms are bonded Hs donating to:
                for bonded_atom_num in atom.bonded_to_atom_num:
                    H = all_data[bonded_atom_num]
                    if H.mass < 1.1 and H.nearest_atom != None:
                        X = H.nearest_atom
                        if X.mass > 1.1:
                            if X.atom_num in shell_list_atom_nums:
                                for rank_num in shell_list_ranked:
                                    rank = rank_num[0]
                                    num = rank_num[1]
                                    if num == X.atom_num:
                                        #treat a broken HB as a SD for now
                                        if (rank, num) not in donor_list:
                                            donor_list.append((rank, num))
                                            break 
                                    else:
                                        continue
                            #check donating inside RAD shell
                            elif X.atom_num not in shell_list_atom_nums \
                                    and X.resname == atom.resname:
                                        donor_list.append(('DX', 
                                            X.atom_num))
                                        logging.warning('DX', atom.atom_name, 
                                                X.atom_name)
                            elif X.atom_num not in shell_list_atom_nums \
                                    and X.resname != atom.resname:
                                        donor_list.append(('DS', 
                                            X.atom_num))
                                        logging.warning('DS', atom.atom_name,
                                                X.atom_name)
                            else:
                                continue

                        #### dealing with H-H interactions.
                        elif X.mass < 1.1:
                            UA_X = None
                            for bonded_atom_num2 in X.bonded_to_atom_num:
                                UA_X = all_data[bonded_atom_num2]
                                if UA_X.mass > 1.1 \
                                        and UA_X.atom_num in \
                                        shell_list_atom_nums:
                                    for rank_num in shell_list_ranked:
                                        rank = rank_num[0]
                                        num = rank_num[1]
                                        if num == UA_X.atom_num:
                                            #treat a H-H interaction as SD and
                                            #as unique to H-X interaction.
                                            if ('H', X.atom_num) \
                                                    not in donor_list:
                                                donor_list.append(('H', 
                                                    X.atom_num))
                                            else:
                                                continue
                                        else:
                                            continue
                                elif UA_X.mass > 1.1 \
                                        and UA_X.atom_num not in \
                                        shell_list_atom_nums:
                                    donor_list.append(('DH', 
                                        X.atom_num))
                                    logging.warning('DH', atom.atom_name,
                                            UA_X.atom_name)
                                else:
                                    continue
                                           

            shell_list_ranked.sort(key=lambda x: x[0])
            acceptor_list.sort(key=lambda x: x[0])
            donor_list.sort(key=lambda x: x[0])


            shell = [i[0] for i in shell_list_ranked]
            acceptors = [i[0] for i in acceptor_list]
            donors = [i[0] for i in donor_list]

            atom.RAD_shell_ranked = (shell, acceptors, donors)

            ### create a list for RAD neighbours and relative 
            ### distances (QQ/r^2)

            RAD_nAtoms = []
            for neighbour in atom.RAD:
                RAD_nAtoms.append([neighbour.atom_name, 
                    neighbour.resid, neighbour.resname])

            atom.RAD_nAtoms = RAD_nAtoms







import numpy as nmp
import sys

from CodeEntropy.FunctionCollection import Utils, CustomFunctions
from CodeEntropy.FunctionCollection import UnitsAndConversions as CONST

class BeadCollection(object):
	""" 
	A class that containes all the properties of 
	the beads/particles that comprise it.
	Properties include values of local vectors and a connection to
	the main data container.
	
	A name is also provided while initializaion
	to be used for writing various outputs related 
	to it.
	"""
	
	def __init__(self, arg_descriptor, arg_hostDataContainer, **kwargs):
		self.name = arg_descriptor
		self.hostDataContainer = arg_hostDataContainer
		self.listOfBeads = [] 
		self.indexPairList = None

		self.subMatrixDict = {"FF" : None,
							  "TT" : None}
		
		# eigen values, frequencues and entropy per DOF
		self.lambdaFreqEntropyDict = {"FF" : None,
									  "TT" : None}
	#END	 

	def assign_attribute(self, arg_attrName, arg_attrValue):
		setattr(self, arg_attrName, arg_attrValue)
		return
	
	def reinitialize_matrices(self):
		N = len(self.listOfBeads)
		if N == 0:
			raise ValueError('There should be at least 1 bead in the system.')
		
		self.indexPairList = [(i,j) for i in range(N) for j in range(N)]
		
		for mKey in self.subMatrixDict.keys():
			self.subMatrixDict[mKey] = nmp.zeros( (N, N, 3, 3) )
			
		return
	#END	 
	
	def generate_quadrant(self, arg_pairString, arg_filterZeros):
		""" For an input sub matrix of 4 dimension where 
		each i-j element is another 3x3 matrix (outer product),
		it used nmp.block to return a 3N x 3N matrix which is called 'Quadrant'

		Sometimes there can be a row or column that is entirely zero. If arg_filterZeros is True, then they must be removed"""

		subMatrix = self.subMatrixDict[arg_pairString]
		nRowsMajor, nColsMajor, nRowsMinor, nColsMinor = nmp.shape(subMatrix)
		
		quadrant = nmp.block([   [  subMatrix[i,j] for j in range(nColsMajor)  ] for i in range(nRowsMajor)   ] )

		if arg_filterZeros:			 
			quadrant = self.filter_zero_rows_columns(quadrant)

		return quadrant
	#END	 

	def filter_zero_rows_columns(self, arg_matrix):
		#record the initial size
		initShape = nmp.shape(arg_matrix)

		zeroIdx = list(filter(lambda row : nmp.all(nmp.isclose(arg_matrix[row,:] , 0.0)) , nmp.arange(nmp.shape(arg_matrix)[0])))
		allIdx = nmp.ones((nmp.shape(arg_matrix)[0]), dtype=bool)
		allIdx[zeroIdx] = False 
		arg_matrix = arg_matrix[allIdx,:]

		allIdx = nmp.ones((nmp.shape(arg_matrix)[1]), dtype=bool)
		zeroIdx = list(filter(lambda col : nmp.all(nmp.isclose(arg_matrix[:,col] , 0.0)) , nmp.arange(nmp.shape(arg_matrix)[1])))
		allIdx[zeroIdx] = False
		arg_matrix = arg_matrix[:,allIdx]

		# get the final shape
		finalShape = nmp.shape(arg_matrix)

		if initShape != finalShape:
			Utils.printflush('A shape change has occured ({},{}) -> ({}, {})'.format(*initShape, *finalShape))

		return arg_matrix
	#END	 

	def update_subMatrix(self, arg_pairString, arg_verbose):
		"""Fill in the matrix with correlation data 
		corresponding to the value for the input arg_pairString (key)"""
	
		if len(self.indexPairList) == 0 or len(self.listOfBeads) == 0:
			raise ValueError("Index Pair List or the beads representing the molecule is not defined")
	
		N = len(self.listOfBeads)
		numFrames = len(self.hostDataContainer.trajSnapshots)
		self.subMatrixDict[arg_pairString] = nmp.zeros( (N, N, 3, 3) )
		
		for i,j in self.indexPairList:
			if i <= j:
				if arg_verbose >= 5:
					Utils.printflush('[{:>5},{:>5}]'.format(i,j), end = ' ')

				iBead = self.listOfBeads[i]
				jBead = self.listOfBeads[j]

				if arg_pairString == 'FF':
					self.subMatrixDict["FF"][i][j] = self.generate_block_matrix(arg_bead_i = iBead, arg_bead_j = jBead, arg_pairString="FF", arg_frameList = nmp.arange(numFrames))
					self.subMatrixDict["FF"][j][i] = self.subMatrixDict["FF"][i][j].T

				elif arg_pairString == "TT":
					self.subMatrixDict["TT"][i][j] = self.generate_block_matrix(arg_bead_i = iBead, arg_bead_j = jBead, arg_pairString="TT", arg_frameList = nmp.arange(numFrames))
					self.subMatrixDict["TT"][j][i] = self.subMatrixDict["TT"][i][j].T
		if arg_verbose >= 3:
			Utils.printflush("Finished updating submatrix {}".format(arg_pairString))
		return
	#END	 
	
	def generate_block_matrix(self, arg_bead_i, arg_bead_j, arg_pairString, arg_frameList):
		""" 
		For a pair of beads/UAs, it will return a 3x3 matrix for 
		each form of arg_pairString. The matrix and its transpose
		when placed correctly into the bigger force-torque covariance 
		matrix will create a block symmetric matrix in each quadrant.

		For each timestep get a cross matrix for the UA pair and then 
		add them serially. In the end, divide the sum by numFrames to
		obtain the average.

		Accepted arg_pairString : ['FF', 'TT']
		"""

		if not arg_pairString.upper() in self.subMatrixDict.keys():
			raise ValueError("Invalid pairString {}. Should be one of these : 'FF', 'TT' ".format(arg_pairString) )

		numFrames = len(self.hostDataContainer.trajSnapshots)

		if arg_pairString.upper() == "FF":
			ffMatrix_ij = nmp.zeros( (3,3) )
			for iFrame in arg_frameList:
				outerProdMatrix = nmp.outer(arg_bead_i.totalWeightedForces[iFrame], arg_bead_j.totalWeightedForces[iFrame])
				ffMatrix_ij = nmp.add(ffMatrix_ij, outerProdMatrix)

			ffMatrix_ij /= numFrames
			return ffMatrix_ij

		if arg_pairString.upper() == "TT":
			ttMatrix_ij = nmp.zeros( (3,3) )
			for iFrame in arg_frameList:
				outerProdMatrix = nmp.outer(arg_bead_i.totalWeightedTorques[iFrame,:], arg_bead_j.totalWeightedTorques[iFrame,:])
				ttMatrix_ij = nmp.add(ttMatrix_ij, outerProdMatrix)

			ttMatrix_ij /= numFrames
			return ttMatrix_ij
	#END

	def get_name(self):
		"""
		Returns the name/descriptor string of this bead collection.
		"""
		return self.name
	#END

	def get_bead_names(self):
		"""
		Returns a string with the identifying `atom names` of the individual comprising beads.
		"""
		retStr = ""
		for iBead in self.listOfBeads:
			retStr = "{} {}".format(retStr, iBead.beadName)

		return retStr
	#END

	def get_bead_resnames(self):
		"""
		Returns a string with the identifying `res names` of the individual comprising beads.
		"""
		retStr = ""
		for iBead in self.listOfBeads:
			retStr = "{} {}".format(retStr, iBead.beadResn)

		return retStr
	#END

	def get_bead_resids(self):
		"""
		Returns a string with the identifying `residue ids` of the individual comprising beads.
		"""
		retStr = ""
		for iBead in self.listOfBeads:
			retStr = "{} {}".format(retStr, iBead.beadResi)

		return retStr
	#END

	def get_bead_chids(self):
		"""
		Returns a string with the identifying `chain ids` of the individual comprising beads.
		"""
		retStr = ""
		for iBead in self.listOfBeads:
			retStr = "{} {}".format(retStr, iBead.beadChain)

		return retStr
	#END

	def get_bead_positions(self):
		"""
		Returns a string with the coordinates (representative) of the individual comprising beads.
		"""
		retStr = ""
		for iBead in self.listOfBeads:
			bx, by, bz = iBead.position
			retStr = "{} {:.3f} {:.3f} {:.3f}".format(retStr, bx, by, bz)

		return retStr
	#END 

	def write_nmd_file(self, arg_nmdfile, arg_spectrum, arg_wfac):
		"""
		write a NMD format file for visualization of the eigen modes in VMD (NMWIZ pugin).
		"""
		Utils.printOut(arg_nmdfile, "title {}".format(self.get_name()))
		Utils.printOut(arg_nmdfile, "names {}".format(self.get_bead_names()))
		Utils.printOut(arg_nmdfile, "resnames {}".format(self.get_bead_resnames()))
		Utils.printOut(arg_nmdfile, "chids {}".format(self.get_bead_chids()))
		Utils.printOut(arg_nmdfile, "resnums {}".format(self.get_bead_resids()))
		Utils.printOut(arg_nmdfile, "coordinates {}".format(self.get_bead_positions()))
		for mode in arg_spectrum:
			Utils.printOut(arg_nmdfile, "# *** {} {:.3f}".format(mode.modeIdx, mode.modeFreq/CONST.C_LIGHT))
			Utils.printOut(arg_nmdfile, mode.get_line_nmd(arg_wfac = arg_wfac))

	#END
		

# END
	


class Bead():
	"""
	It is a collection of atoms and it will contain 
	the total forces and torques (weighted) for each of its
	atoms per frame. It has a hostDataContainer from where 
	it fetches some of the information about its constituent atoms.
	"""

	def __init__(self, arg_atomList, arg_numFrames, arg_hostDataContainer\
					 , arg_beadName = "", arg_beadResn = "", arg_beadResi = ""\
					 , arg_beadChid = "", **kwargs):
		
		# set of atom indices that form this bead
		self.atomList = arg_atomList

		# a 3D coordinate representing its position in space
		self.position = None

		# dummy information about beads
		self.beadName = arg_beadName
		self.beadResn = arg_beadResn
		self.beadResi = arg_beadResi
		self.beadChain = arg_beadChid		
		
		#mass weighted local forces	
		self.totalWeightedForces = nmp.zeros( (arg_numFrames,3) )	 
		
		#mass weighted local torques
		self.totalWeightedTorques = nmp.zeros( (arg_numFrames,3) )   
		
		# data container
		self.hostDataContainer = arg_hostDataContainer
	# END

	def reset_totalWeightedVectors(self, arg_shape):
		self.totalWeightedForces = nmp.zeros( arg_shape )
		self.totalWeightedTorques = nmp.zeros( arg_shape )
		return	
	# END

	def get_num_atoms(self):
		return len(self.atomList)
	#ND

	def get_total_mass(self):
		totalMass = 0.
		for aid in self.atomList:
			iMass = self.hostDataContainer.universe.atoms.masses[aid]
			totalMass += iMass

		return totalMass
	# END

	def get_center_of_mass_lab(self, arg_frame):
		""" 
		compute and return the center of mass of 
		the input atoms in the lab frame
		"""
		
		com = nmp.zeros((3))

		totalMass = nmp.sum(nmp.asarray([self.hostDataContainer.universe.atoms.masses[iAtom] for iAtom in self.atomList]))

		for iAtom in self.atomList:
			iMassCoord = self.hostDataContainer.universe.atoms.masses[iAtom] * self.hostDataContainer._labCoords[arg_frame][iAtom]
			com = com + iMassCoord

		com /= totalMass
		return com
	# END

	def get_center_of_mass_local(self, arg_frame):
		""" 
		compute and return the center of mass of 
		the input atoms in the local frame
		"""
		
		com = nmp.zeros((3))

		totalMass = nmp.sum(nmp.asarray([self.hostDataContainer.universe.atoms.masses[iAtom] for iAtom in self.atomList]))

		for iAtom in self.atomList:
			iMassCoord = self.hostDataContainer.universe.atoms.masses[iAtom] * self.hostDataContainer.localCoords[arg_frame][iAtom]
			com = com + iMassCoord

		com /= totalMass
		return com
	# END

	def get_moment_of_inertia_tensor_local(self, arg_localBasis, arg_frame):
		""" 
		return the 3 x 3 moment of inertia tensor for 
		the body defined by the collection of atoms in 
		the bead. The MOIs are computed in the input basis. 
		The lab coordinates of the atoms will be transformed 
		into this basis and then the moment would be computed. 

		NB: Moment of inertia tensor expression taken from MDanalysis website.
		It agrees with the algebric expression provided elsewhere
		"""

		#check that the input basis is of shape 4,3
		try:
			assert(nmp.shape(arg_localBasis) == (4,3))
		except:
			Utils.printflush('The dimensions of the input basis must be 4 x 3')

		# now compute the local coords for each atom in the input list in that frame
		# store them
		atomLocalCoords = nmp.ndarray( (len(self.atomList),3) )
		atomMasses = nmp.zeros( (len(self.atomList)) )

		atIdx = 0
		for iAtom in self.atomList:
			atomLocalCoords[atIdx] = self.hostDataContainer._labCoords[arg_frame][iAtom] - arg_localBasis[-1,]
			atomLocalCoords[atIdx] = arg_localBasis[0:3,] @ atomLocalCoords[atIdx]

			atomMasses[atIdx] = self.hostDataContainer.universe.atoms.masses[iAtom]

			atIdx += 1


		# generate the  MOI tensor
		tensor = nmp.zeros((3,3))

		sumDyadicCoordBasis = nmp.zeros((3,3))
		for iAxis in range(3):
			sumDyadicCoordBasis = nmp.add(sumDyadicCoordBasis, nmp.outer(arg_localBasis[iAxis,], arg_localBasis[iAxis,]))

		for atIdx in range(len(self.atomList)):
			iMass = atomMasses[atIdx]
			iLocalCoord = atomLocalCoords[atIdx]

			iDyadicLocalCoord = nmp.outer(iLocalCoord, iLocalCoord)
			
			iMatrix = nmp.dot(iLocalCoord,iLocalCoord) * sumDyadicCoordBasis
			iMatrix = nmp.subtract(iMatrix,iDyadicLocalCoord)
			iMatrix = iMass * iMatrix
			
			tensor = nmp.add(tensor, iMatrix)


		return tensor
#END	

	def print_atomList(self):
		""" Prints the atom indices in a particular fashion"""
		baseMol = self.hostDataContainer.universe
		Utils.printflush('[{}]'.format(len(self.atomList)), end = ' : ')
		for iAtom in self.atomList:
			if iAtom in baseMol.select_atoms("name H*").indices:
				Utils.printflush('{}'.format(iAtom),end = ' ')
			else:
				Utils.printflush('({})'.format(iAtom), end = ' ')
		Utils.printflush('')
		return

#END				




import numpy as nmp
import sys

from CodeEntropy.FunctionCollection import CustomFunctions, Utils

def write_file(arg_filename):
	"""
	Open a file stream for writing. It will overwrite any exisitng file
	by that name.
	"""
	fwrite = open(arg_filename,"w")
	fwrite.close()
	return 
#END

def append_file(arg_filename):
	"""
	Open a file stream for appending.
	"""
	fapp = open(arg_filename,"a")
	fapp.close()
#END

def write_atomic_forces( arg_baseMolecule, arg_hostDataContainer, arg_outFile ):
	""" function that prints the 
	1. average
	2. SD
	of the forces felt by each atom
	"""

	numFrames = len(arg_hostDataContainer.trajSnapshots)
	for iAtom in arg_baseMolecule.atomIndexArray:
		if not arg_baseMolecule.isHydrogenArray[iAtom]:
			iAtomname = arg_baseMolecule.atomNameArray[iAtom]
			iResid  = arg_baseMolecule.atomResidueIdxArray[iAtom]
			iResname = arg_baseMolecule.resnameArray[iResid]
			iChain = "X"

			iForces = arg_hostDataContainer._labForces[:,iAtom]
			try:
				assert(nmp.shape(iForces) == (numFrames, 3))
			except:
				Utils.printflush('Requested shape ', nmp.shape(iForces), 'does not allow proper broadcasting of the array of forces from the main data containter.')

			#coordinates
			iAvgX = nmp.mean(nmp.asarray([arg_hostDataContainer._labCoords[:, iAtom, 0]]))
			iAvgY = nmp.mean(nmp.asarray([arg_hostDataContainer._labCoords[:, iAtom, 1]]))
			iAvgZ = nmp.mean(nmp.asarray([arg_hostDataContainer._labCoords[:, iAtom, 2]]))
			
			# forces
			iForceMagnitudes = nmp.asarray([ nmp.linalg.norm(iForces[i]) for i in range(numFrames)])

			# average force
			iAvgForce = nmp.mean(iForceMagnitudes)

			# std dev of forces
			iSDevForce = nmp.std(iForceMagnitudes)

			#pdb format string
			atomStr = '{:<6s}{:>5d} {:>4s} {:3s} {:1s}{:>4d}    {:>8.3f}{:>8.3f}{:>8.3f}{:>6.2f}{:>6.2f}'.format("ATOM", iAtom, iAtomname, iResname, iChain, iResid, iAvgX, iAvgY, iAvgZ, iAvgForce, iSDevForce )
			Utils.printOut(arg_outFile, atomStr)

	return
#END

def write_a_matrix(arg_matrix, arg_descriptor, arg_outFile):
	""" This function writes a matrix to an ASCII outputfile. The elemetns are
	written in the format "%5d %5d %15.5f" in each line."""

	# fetch the shape (it must be two dimesnional)
	matShape = nmp.shape(arg_matrix)
	try:
		assert(len(matShape) == 2)
	except:
		print("A 2D matrix is expected. Received a matrix with {} dimensions instead. Will not write anything therefore.".format(len(matShape))) 
		return

	# write
	Utils.printOut(arg_outFile, "***MATRIX: {}".format(arg_descriptor))
	
	nrow, ncol = matShape
	Utils.printOut(arg_outFile, "***MATRIX: SHAPE {} {}".format(*matShape))
	Utils.printOut(arg_outFile, "***MATRIX: TRACE {:<15.5f}".format(nmp.trace(arg_matrix)))
	for iRow in range(nrow):
		for jCol in range(ncol):
			Utils.printOut(arg_outFile, "{:>5d}{:>5d}{:>15.5f}".format(iRow, jCol, arg_matrix[iRow][jCol]))
	Utils.printOut(arg_outFile, "***MATRIX: END")

	return

	
